//
// Display and keyboard functionality
// (c) 9.10.2016 Ingo D. Rullhusen
//

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>


#include "diskey.h"


volatile uint8_t scan_state = 0;
uint8_t SEGMENT[6] = { 0xff, 0x86, 0x7f, 0xff, 0xff, 0xff };
volatile uint8_t key_matrix[6];

volatile uint8_t KEY = K_NONE;
volatile uint8_t MODE = M_Error;
volatile uint8_t AUTOSCAN;

volatile uint8_t diskey_set_pll;
uint8_t tuneFreqR[8];


void diskey_init(void)
{
                 // WGM22:0 : 2 = CTC mode
   TCCR2A = 0x2; // no port connexion WGM21:0
   TCCR2B = 0x6; // WGM22 clk/256 => 15.625 kHz => 64 µs

   OCR2A = 0x0a; // f = 15.625 kHz / (OCR2A + 1)
   OCR2B = 0x08;

   TIMSK2 = 0x6; // OCF2A
// TIFR2 interrupt flag 

// ports
   DDRA = 0xff; // DIGIT OUT & debug & LATCH OUT

// pLED
   DDRC |= 0xf0; // BCDOUT
   PORTC |= 0xf0;

// key matrix
   DDRD &= ~(0x3f);
   PORTD &= ~(0x3f); // no pull-up
//   PORTD |= 0x3f; // pull-up

   DDRF = 0xff; // SEGMENT OUT

}

void diskey_pLED( uint8_t led )
{
   PORTC = (led<<4) | (PORTC&0x0f);
}
   
void diskey_sLED( uint8_t led )
{
   if (led) { SEGMENT[0] &= ~(0x20); }
   else     { SEGMENT[0] |=   0x20;  }
}


ISR(TIMER2_COMPB_vect)
{
   if ( scan_state > 0 ) { key_matrix[scan_state-1] = 0x3f & ~PIND; }
   PORTF = 0xff;
   PORTA |= 0x3f;
}

void update_key(void)
{
   switch( key_matrix[0] )
   {
      case 0x20: { MODE = M_ON;     break; }
      case 0x10: { MODE = M_OFF;    break; }
      case 0x08: { MODE = M_CSET;   break; }
      case 0x04: { MODE = M_CHECK;  break; }
      case 0x02: { MODE = M_TSET;   break; }
      case 0x01: { MODE = M_CANCEL; break; }

      case 0x11: { MODE = M_RESET;  break; }
      case 0x30: { MODE = M_MSG;    break; }

      default  : { MODE = M_Error;  break; }
   }

   AUTOSCAN = key_matrix[4] & 0x10;

   if ( key_matrix[3] == 0x0c ) { KEY = K_RESET; return; }

   uint8_t nzb = 0;
   if ( key_matrix[1]      ) { nzb |= 0x10; }
   if ( key_matrix[2]      ) { nzb |= 0x08; }
   if ( key_matrix[3]      ) { nzb |= 0x04; }
   if ( key_matrix[4]&0xef ) { nzb |= 0x02; }
   if ( key_matrix[5]      ) { nzb |= 0x01; }

   switch( nzb )
   {
      case 0x10:
      {
         switch( key_matrix[1] )
         {
            case 0x08: { KEY = K_1; break; }
            case 0x04: { KEY = K_2; break; }
            case 0x02: { KEY = K_3; break; }
            case 0x01: { KEY = K_4; break; }
            default  : { KEY = K_NONE; break; }
         }
         break;
      }
      case 0x08:
      {
         switch( key_matrix[2] )
         {
            case 0x08: { KEY = K_5; break; }
            case 0x04: { KEY = K_6; break; }
            case 0x02: { KEY = K_7; break; }
            case 0x01: { KEY = K_8; break; }
            default  : { KEY = K_NONE; break; }
         }
         break;
      }
      case 0x04:
      {
         switch( key_matrix[3] )
         {
            case 0x08: { KEY = K_9_FM; break; }
            case 0x04: { KEY = K_0_MW; break; }
            default  : { KEY = K_NONE; break; }
         }
         break;
      }
      case 0x02:
      {
         switch( key_matrix[4]&0xef )
         {
            case 0x02: { KEY = K_RIGHT; break; }
            case 0x01: { KEY = K_LEFT;  break; }
            default  : { KEY = K_NONE; break; }
         }
         break;
      }
      case 0x01:
      {
         switch( key_matrix[5] )
         {
            case 0x04: { KEY = K_STORE;  break; }
            case 0x01: { KEY = K_PWROFF; break; }
            default  : { KEY = K_NONE; break; }
         }
         break;
      }
      default  : { KEY = K_NONE; break; }
   }

}

void tune_pll(void)
{
   PORTA &= 0x7f;
   PORTF &= 0xf0;
   for ( uint8_t i=0; i<8; ++i )
   {
      PORTF = ( i + 1 );
      _delay_us(10);
      PORTA |= 0x80;
      _delay_us(10);
      PORTF = tuneFreqR[i];
      _delay_us(10);
      PORTA &= 0x7f;
      _delay_us(10);
   }
}

ISR(TIMER2_COMPA_vect)
{
   switch ( scan_state )
   {
      case 0:
      {
         PORTA = 0x3e;
         PORTF = SEGMENT[0];
         OCR2A = 0x0a;
         OCR2B = 0x08;
         ++scan_state;
         break;
      }
      case 1:
      {
         PORTA = 0x3d;
         PORTF = SEGMENT[1];
         ++scan_state;
         break;
      }
      case 2:
      {
         PORTA = 0x3b;
         PORTF = SEGMENT[2];
         ++scan_state;
         break;
      }
      case 3:
      {
         PORTA = 0x37;
         PORTF = SEGMENT[3];
         ++scan_state;
         break;
      }
      case 4:
      {
         PORTA = 0x2f;
         PORTF = SEGMENT[4];
         ++scan_state;
         break;
      }
      case 5:
      {
         PORTA = 0x1f;
         PORTF = SEGMENT[5];
         ++scan_state;
         break;
      }
      case 6:
      {
// key matrix -> key
         update_key();

// pll data
         if ( diskey_set_pll )
         {
            tune_pll();
            diskey_set_pll = PLL_TUNED;
         }

         OCR2A = 0x36;
         OCR2B = 0x34;
         scan_state = 0;
         break;
      }
      default:
      { scan_state = 0; OCR2A = 0x0a; OCR2B = 0x08; break; }
   }

}


static uint8_t PROGMEM BCDlut[] = {
   0xc0, 0xf9, 0xa4, 0xb0, 0x99, 0x92, 0x82, 0xf8, //   0
   0x80, 0x90, 0x88, 0x83, 0xc6, 0xa1, 0x86, 0x8e, //   8
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, //  16
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, //  24
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, //  32
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, //  40
   0xc0, 0xf9, 0xa4, 0xb0, 0x99, 0x92, 0x82, 0xf8, //  48
   0x80, 0x90, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, //  56
   0xff, 0x88, 0xff, 0xc6, 0xff, 0x86, 0x8e, 0xc2, //  64
   0xff, 0xff, 0xff, 0xff, 0xc7, 0xff, 0xff, 0xff, //  72
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, //  80
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf7, //  88
   0xff, 0xff, 0x83, 0xff, 0xa1, 0xff, 0xff, 0xff, //  96
   0x8b, 0xe7, 0xff, 0xff, 0xff, 0xff, 0xab, 0xa3, // 104
   0xff, 0xff, 0xaf, 0x93, 0x87, 0xe3, 0xff, 0xff, // 112
   0xff, 0x91, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 120
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 128
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 136
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 144
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 152
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 160
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 168
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 176
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 184
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 192
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 200
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 208
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 216
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 224
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 232
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, // 240
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff  // 248
                                  };

uint8_t diskey_hex2seg( uint8_t c )
{
   return( pgm_read_byte(BCDlut+c) );
}

void diskey_time2seg( bcd_time *clk )
{
   SEGMENT[1] = ( clk->H == 0 ? 0xff : diskey_hex2seg( clk->H ) );
   SEGMENT[2] = diskey_hex2seg( clk->h ) & 0x7f;
   SEGMENT[3] = diskey_hex2seg( clk->M );
   SEGMENT[4] = diskey_hex2seg( clk->m );
   SEGMENT[5] = 0xff;
   SEGMENT[0] |= 0xdf;
}

void diskey_fm2seg( freq10k_fm f )
{
   uint8_t bcd;

   SEGMENT[0] = (SEGMENT[0]&0xf6) | 0x01;

   for ( bcd=0; f>9999; ++bcd, f-=10000 ) {}
   SEGMENT[1] = ( bcd==0 ? 0xff : diskey_hex2seg(bcd) );

   for ( bcd=0; f>999; ++bcd, f-=1000 ) {}
   SEGMENT[2] = diskey_hex2seg(bcd);

   for ( bcd=0; f>99; ++bcd, f-=100 ) {}  
   SEGMENT[3] = diskey_hex2seg(bcd) & 0x7f;

   for ( bcd=0; f>9; ++bcd, f-=10 ) {}
   SEGMENT[4] = diskey_hex2seg(bcd);

   SEGMENT[5] = diskey_hex2seg(f);
}

void diskey_am2seg( freq1k_am f )
{
   uint8_t bcd;

   SEGMENT[0] = (SEGMENT[0]&0xf6) | 0x08;

   for ( bcd=0; f>999; ++bcd, f-=1000 ) {}
   SEGMENT[1] = ( bcd==0 ? 0xff : diskey_hex2seg(bcd) );

   for ( bcd=0; f>99; ++bcd, f-=100 ) {}
   SEGMENT[2] = diskey_hex2seg(bcd);

   for ( bcd=0; f>9; ++bcd, f-=10 ) {}  
   SEGMENT[3] = diskey_hex2seg(bcd);

   SEGMENT[4] = diskey_hex2seg(f);

   SEGMENT[5] = 0xff;
}

void diskey_str2seg( const char *PSTRs, const uint8_t p2, const uint8_t p3 )
{
   uint8_t segno = 1;
   char c;

   SEGMENT[0] |= 0x09;

   for ( ; segno<6; ++segno )
   {
      c = pgm_read_byte(PSTRs++);
      if (!c) break;
      if ( (segno == 2 && p2) || (segno == 3 && p3) )
      { SEGMENT[segno] = diskey_hex2seg(c) & 0x7f; }
      else
      { SEGMENT[segno] = diskey_hex2seg(c); }
   }

   for ( ; segno<6; ++segno )
   {
      if ( (segno == 2 && p2) || (segno == 3 && p3) )
      { SEGMENT[segno] = 0x7f; }
      else
      { SEGMENT[segno] = 0xff; }
   }
}

