//
// Timer functionality
// (c) 8.10.2016 Ingo D. Rullhusen
//

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "timer.h"
#include "diskey.h"


volatile bcd_time CLOCK;

timer PERMANENT, ONCE;


void timer_var_clear( timer *T )
{
   T->on.H = T->on.h = T->on.M = T->on.m = 0xff;
   T->off.H = T->off.h = T->off.M = T->off.m = 0xff;
   T->band = BAND_OFF;
   T->station = 0xff;
}


void timer_init(void)
{
// WGM33:0 : 4 = CTC mode
   TCCR3A = 0x00; // no output; WGM31:0
   TCCR3B = 0x0f; // no capture; WGM33:2; in/1 = 250 Hz

   OCR3A = (unsigned int)14999; // f = 1 / min

   TIMSK3 = 0x02; // OCF3A
// TIFR3 interrupt flag 

// T3 input
   DDRC &= 0xf7;
//   PORTC &= 0xf7; // no pull-up
   PORTC |= 0x08; // pull-up

   TCNT3 = 0;
   CLOCK.H = 0xff;
   CLOCK.h = 0xff;
   CLOCK.M = 0xff;
   CLOCK.m = 0xff;

   timer_var_clear( &PERMANENT );
   timer_var_clear( &ONCE );

}

ISR(TIMER3_COMPA_vect)
{
   if ( CLOCK.H > 2 || CLOCK.h > 9 || CLOCK.M > 5 || CLOCK.m > 9 )
   {  // error
      CLOCK.H = 0xff;
      CLOCK.h = 0xff;
      CLOCK.M = 0xff;
      CLOCK.m = 0xff;
      return;
   }

   ++CLOCK.m;
   if ( CLOCK.m > 9 ) { CLOCK.m = 0; ++CLOCK.M; }
   if ( CLOCK.M > 5 ) { CLOCK.M = 0; ++CLOCK.h; }
   if ( CLOCK.H < 2 )
   { if ( CLOCK.h > 9 ) { CLOCK.h = 0; ++CLOCK.H;     } }
   else
   { if ( CLOCK.h > 3 ) { CLOCK.h = 0;   CLOCK.H = 0; } }


}


uint8_t time_input( bcd_time *time )
{
   uint8_t mode_t = MODE;

   time->H = '_';
   time->h = 0xff;
   time->M = 0xff;
   time->m = 0xff;
   diskey_time2seg( time );

   while ( MODE == mode_t )
   {
      if ( KEY == K_0_MW ) { time->H = 0; break; }
      if ( KEY == K_1 ) { time->H = 1; break; }
      if ( KEY == K_2 ) { time->H = 2; break; }
   }
   diskey_time2seg( time );
   while ( MODE == mode_t && KEY != K_NONE ) {}
   if ( MODE != mode_t ) { return(0); }

   time->h = '_';
   diskey_time2seg( time );
   while ( MODE == mode_t )
   {
      if ( KEY == K_0_MW ) { time->h = 0; break; }
      if ( KEY == K_1 ) { time->h = 1; break; }
      if ( KEY == K_2 ) { time->h = 2; break; }
      if ( KEY == K_3 ) { time->h = 3; break; }
      if ( time->H < 2 )
      {
         if ( KEY == K_4 ) { time->h = 4; break; }
         if ( KEY == K_5 ) { time->h = 5; break; }
         if ( KEY == K_6 ) { time->h = 6; break; }
         if ( KEY == K_7 ) { time->h = 7; break; }
         if ( KEY == K_8 ) { time->h = 8; break; }    
         if ( KEY == K_9_FM ) { time->h = 9; break; }
      }
   }
   diskey_time2seg( time );
   while ( MODE == mode_t && KEY != K_NONE ) {}
   if ( MODE != mode_t ) { return(0); }

   time->M = '_';
   diskey_time2seg( time );
   while ( MODE == mode_t )
   {
      if ( KEY == K_0_MW ) { time->M = 0; break; }
      if ( KEY == K_1 ) { time->M = 1; break; }
      if ( KEY == K_2 ) { time->M = 2; break; }
      if ( KEY == K_3 ) { time->M = 3; break; }
      if ( KEY == K_4 ) { time->M = 4; break; }
      if ( KEY == K_5 ) { time->M = 5; break; }
   }
   diskey_time2seg( time );
   while ( MODE == mode_t && KEY != K_NONE ) {}
   if ( MODE != mode_t ) { return(0); }

   time->m = '_';
   diskey_time2seg( time );
   while ( MODE == mode_t )
   {
      if ( KEY == K_0_MW ) { time->m = 0; break; }
      if ( KEY == K_1 ) { time->m = 1; break; }
      if ( KEY == K_2 ) { time->m = 2; break; }
      if ( KEY == K_3 ) { time->m = 3; break; }
      if ( KEY == K_4 ) { time->m = 4; break; }
      if ( KEY == K_5 ) { time->m = 5; break; }
      if ( KEY == K_6 ) { time->m = 6; break; }
      if ( KEY == K_7 ) { time->m = 7; break; }
      if ( KEY == K_8 ) { time->m = 8; break; }
      if ( KEY == K_9_FM ) { time->m = 9; break; }
   }
   diskey_time2seg( time );

   _delay_ms(1000);

   return(1);
}

void time_set( const bcd_time *clk )
{
   TCNT3 = (uint16_t)0;
   CLOCK = *clk;
}


uint8_t setT( timer *T )
{
   uint8_t led = 0;
   uint8_t k_t;

   while ( KEY != K_NONE && MODE == M_TSET ) {}
   if ( MODE != M_TSET ) { return(0); }

   if ( ! time_input( &T->on ) ) { return(0); }

   for ( int i=0; i<10 && MODE == M_TSET; ++i )
   { _delay_ms(100); }
   if ( MODE != M_TSET ) { return(0); }

   led = 0;
   k_t = KEY;
   while ( k_t!=K_0_MW && k_t!=K_9_FM && MODE == M_TSET )
   {
      set_band( ( led & 0x01 ? BAND_FM : BAND_MW ) );
      for ( int i = 0;
            i < 20 && k_t != K_0_MW && k_t != K_9_FM && MODE == M_TSET;
            ++i )
      { _delay_ms(10); k_t = KEY; }
      ++led;
   }

   if ( MODE != M_TSET ) { return(0); }
   T->band = ( k_t == K_0_MW ? BAND_MW : BAND_FM );
   set_band( T->band );

   led = 0;
   k_t = KEY;
   while ( (k_t<K_1 || k_t>K_8) && MODE == M_TSET )
   {
      diskey_pLED( led & 0x07 );
      for ( int i = 0;
            i < 20 && (k_t<K_1 || k_t>K_8) && MODE == M_TSET;
            ++i )
      { _delay_ms(10); k_t = KEY; }
      ++led;
   }

   if ( MODE != M_TSET ) { return(0); }
   T->station = k_t - K_1;

   _delay_ms(1000);

   return(1);
}

void timer_set(void)
{
   uint8_t led;
   uint8_t k_t;
   timer T;

   diskey_str2seg( PSTR("     "), 1, 0 );
   timer_var_clear(&T);

   while ( KEY != K_NONE && MODE == M_TSET ) {}
   if ( MODE != M_TSET ) { return; }

   led = 0;
   k_t = KEY;
   while ( (k_t<K_1 || k_t>K_4) && MODE == M_TSET )
   {
      diskey_pLED( led & 0x03 );
      for ( int i=0; i<20 && (k_t<K_1 || k_t>K_4) && MODE == M_TSET; ++i )
      { _delay_ms(10); k_t = KEY; }
      ++led;
   }

   while ( KEY != K_NONE && MODE == M_TSET ) {}
   if ( MODE != M_TSET ) { return; }
   diskey_pLED( k_t - K_1 );

   switch (k_t)
   {
      case K_1:
      {
         if ( setT(&T) )
         {
            PERMANENT.on = T.on;
            PERMANENT.band = T.band;
            PERMANENT.station = T.station;
         }
         break;
      }
      case K_2:
      {
         if ( time_input( &T.off ) ) { PERMANENT.off = T.off; }
         break;
      }
      case K_3:
      {
         if ( setT(&T) )
         {
            ONCE.on = T.on;
            ONCE.band = T.band;
            ONCE.station = T.station;
         }
         break;
      }
      case K_4:
      {
         if ( time_input( &T.off ) ) { ONCE.off = T.off; }
         break;
      }
   }

}

void timer_cancel(void)
{
   static uint8_t led = 0;

   diskey_str2seg( PSTR("CLEAr"), 0, 0 );
   diskey_pLED(0xff);

   switch (KEY)
   {
      case K_1:
      {
         PERMANENT.on.H = PERMANENT.on.h
            = PERMANENT.on.M = PERMANENT.on.m = 0xff;
         PERMANENT.band = BAND_OFF;
         PERMANENT.station = 0xff;
         break;
      }
      case K_2:
      {
         PERMANENT.off.H = PERMANENT.off.h
            = PERMANENT.off.M = PERMANENT.off.m = 0xff;
         break;
      }
      case K_3:
      {
         ONCE.on.H = ONCE.on.h = ONCE.on.M = ONCE.on.m = 0xff;
         ONCE.band = BAND_OFF;
         ONCE.station = 0xff;
         break;
      }
      case K_4:
      {
         ONCE.off.H = ONCE.off.h = ONCE.off.M = ONCE.off.m = 0xff;
         break;
      }
   }

   if ( (led&0x03) == 0 && PERMANENT.on.H  < 3 ) { diskey_pLED(0); }
   if ( (led&0x03) == 1 && PERMANENT.off.H < 3 ) { diskey_pLED(1); }
   if ( (led&0x03) == 2 && ONCE.on.H  < 3 )      { diskey_pLED(2); }
   if ( (led&0x03) == 3 && ONCE.off.H < 3 )      { diskey_pLED(3); }

   ++led;
}

void timer_check(void)
{

   diskey_str2seg( PSTR("     "), 1, 0 );

   while ( MODE == M_CHECK )
   {

      switch ( KEY )
      {
         case K_1:
         {
            diskey_time2seg( &PERMANENT.on );
            set_band( PERMANENT.band );
            diskey_pLED( PERMANENT.station );
            break;
         }
         case K_2:
         {
            diskey_time2seg( &PERMANENT.off );
            set_band( BAND_OFF );
            diskey_pLED( 0xff );
            break;
         }
         case K_3:
         {
            diskey_time2seg( &ONCE.on );
            set_band( ONCE.band );
            diskey_pLED( ONCE.station );
            break;
         }
         case K_4:
         {
            diskey_time2seg( &ONCE.off );
            set_band( BAND_OFF );
            diskey_pLED( 0xff );
            break;
         }
      }

   }

}

